# -*- encoding: utf-8 -*-
#
# Copyright © 2014 eNovance
#
# Authors: Mehdi Abaakouk <mehdi.abaakouk@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import contextlib
import ctypes
import errno
import time

from oslo_config import cfg
from oslo_utils import importutils

from gnocchi import storage
from gnocchi.storage import _carbonara

# NOTE(sileht): rados module is not available on pypi
rados = importutils.try_import('rados')

OPTS = [
    cfg.StrOpt('ceph_pool',
               default='gnocchi',
               help='Ceph pool name to use.'),
    cfg.StrOpt('ceph_username',
               default=None,
               help='Ceph username (ie: client.admin).'),
    cfg.StrOpt('ceph_keyring',
               default=None,
               help='Ceph keyring path.'),
    cfg.StrOpt('ceph_conffile',
               default='/etc/ceph/ceph.conf',
               help='Ceph configuration file.'),
]


class CephStorage(_carbonara.CarbonaraBasedStorage):
    def __init__(self, conf):
        super(CephStorage, self).__init__(conf)
        self.pool = conf.ceph_pool
        options = {}
        if conf.ceph_keyring:
            options['keyring'] = conf.ceph_keyring

        # NOTE(sileht): librados handles reconnection itself,
        # by default if a call timeout (30sec), it raises
        # a rados.Timeout exception, and librados
        # still continues to reconnect on the next call
        self.rados = rados.Rados(conffile=conf.ceph_conffile,
                                 rados_id=conf.ceph_username,
                                 conf=options)
        self.rados.connect()

    @contextlib.contextmanager
    def _lock(self, metric, lock_name):
        # NOTE(sileht): current stable python binding (0.80.X) doesn't
        # have rados_lock_XXX method, so do ourself the call with ctypes
        #
        # https://github.com/ceph/ceph/commit/f5bf75fa4109b6449a88c7ffdce343cf4691a4f9
        # When ^^ is released, we can drop this code and directly use:
        # - ctx.lock_exclusive(name, 'lock', 'gnocchi')
        # - ctx.unlock(name, 'lock', 'gnocchi')
        name = self._get_object_name(metric, lock_name)
        with self._get_ioctx() as ctx:
            while True:
                ret = rados.run_in_thread(
                    ctx.librados.rados_lock_exclusive,
                    (ctx.io, ctypes.c_char_p(name.encode('ascii')),
                     ctypes.c_char_p(b"lock"),
                     ctypes.c_char_p(b"gnocchi"),
                     ctypes.c_char_p(b""), None, ctypes.c_int8(0)))
                if ret in [errno.EBUSY, errno.EEXIST]:
                    time.sleep(0.1)
                elif ret < 0:
                    rados.make_ex(ret, "Error while getting lock of %s" % name)
                else:
                    break
            try:
                yield
            finally:
                ret = rados.run_in_thread(
                    ctx.librados.rados_unlock,
                    (ctx.io, ctypes.c_char_p(name.encode('ascii')),
                     ctypes.c_char_p(b"lock"), ctypes.c_char_p(b"gnocchi")))
                if ret < 0:
                    rados.make_ex(ret,
                                  "Error while releasing lock of %s" % name)

    def _get_ioctx(self):
        return self.rados.open_ioctx(self.pool)

    @staticmethod
    def _get_object_name(metric, lock_name):
        return str("gnocchi_%s_%s" % (metric.id, lock_name))

    @staticmethod
    def _object_exists(ioctx, name):
        try:
            size, mtime = ioctx.stat(name)
            # NOTE(sileht): the object have been created by
            # the lock code
            return size > 0
        except rados.ObjectNotFound:
            return False
        return True

    def _create_metric_container(self, metric):
        name = self._get_object_name(metric, 'container')
        with self._get_ioctx() as ioctx:
            if self._object_exists(ioctx, name):
                raise storage.MetricAlreadyExists(metric)
            else:
                ioctx.write_full(name, "metric created")

    def _store_metric_measures(self, metric, aggregation, data):
        name = self._get_object_name(metric, aggregation)
        with self._get_ioctx() as ioctx:
            ioctx.write_full(name, data)

    def delete_metric(self, metric):
        with self._get_ioctx() as ioctx:
            name = self._get_object_name(metric, 'container')
            try:
                ioctx.remove_object(name)
            except rados.ObjectNotFound:
                raise storage.MetricDoesNotExist(metric)
            for aggregation in metric.archive_policy.aggregation_methods:
                name = self._get_object_name(metric, aggregation)
                try:
                    ioctx.remove_object(name)
                except rados.ObjectNotFound:
                    pass

    def _get_measures(self, metric, aggregation):
        try:
            with self._get_ioctx() as ioctx:
                name = self._get_object_name(metric, aggregation)
                offset = 0
                content = b''
                while True:
                    data = ioctx.read(name, offset=offset)
                    if not data:
                        break
                    content += data
                    offset += len(content)
                if len(content) == 0:
                    # NOTE(sileht: the object have been created by
                    # the lock code
                    raise rados.ObjectNotFound
                else:
                    return content
        except rados.ObjectNotFound:
            name = self._get_object_name(metric, 'container')
            with self._get_ioctx() as ioctx:
                if self._object_exists(ioctx, name):
                    raise storage.AggregationDoesNotExist(metric, aggregation)
                else:
                    raise storage.MetricDoesNotExist(metric)



#
# Copyright 2015 eNovance
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystoneclient.v2_0 import client as ksclient
from oslo_config import cfg

cfg.CONF.import_group('service_credentials', 'ceilometer.service')


def get_keystone_client():
    return ksclient.Client(
        username=cfg.CONF.service_credentials.os_username,
        password=cfg.CONF.service_credentials.os_password,
        tenant_id=cfg.CONF.service_credentials.os_tenant_id,
        tenant_name=cfg.CONF.service_credentials.os_tenant_name,
        cacert=cfg.CONF.service_credentials.os_cacert,
        auth_url=cfg.CONF.service_credentials.os_auth_url,
        region_name=cfg.CONF.service_credentials.os_region_name,
        insecure=cfg.CONF.service_credentials.insecure)



#
# Copyright 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from gnocchi.ceilometer.resources import base


class Identity(base.ResourceBase):
    @staticmethod
    def get_resource_extra_attributes(sample):
        return {}

    @staticmethod
    def get_metrics_names():
        return ['identity.authenticate.success',
                'identity.authenticate.pending',
                'identity.authenticate.failure',
                'identity.user.created',
                'identity.user.deleted',
                'identity.user.updated',
                'identity.group.created',
                'identity.group.deleted',
                'identity.group.updated',
                'identity.role.created',
                'identity.role.deleted',
                'identity.role.updated',
                'identity.project.created',
                'identity.project.deleted',
                'identity.project.updated',
                'identity.trust.created',
                'identity.trust.deleted',
                'identity.role_assignment.created',
                'identity.role_assignment.deleted',
                ]



# -*- encoding: utf-8 -*-
#
# Copyright © 2014-2015 eNovance
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import os

import keystonemiddleware.auth_token
from oslo_log import log
from oslo_policy import policy
from oslo_utils import importutils
import pecan
from pecan import templating
import webob.exc
from werkzeug import serving
from werkzeug import wsgi

from gnocchi import exceptions
from gnocchi import indexer
from gnocchi import json
from gnocchi import service
from gnocchi import storage


LOG = log.getLogger(__name__)


class GnocchiHook(pecan.hooks.PecanHook):

    def __init__(self, storage, indexer, conf):
        self.storage = storage
        self.indexer = indexer
        self.conf = conf
        self.policy_enforcer = policy.Enforcer(conf)

    def on_route(self, state):
        state.request.storage = self.storage
        state.request.indexer = self.indexer
        state.request.conf = self.conf
        state.request.policy_enforcer = self.policy_enforcer


class OsloJSONRenderer(object):
    @staticmethod
    def __init__(*args, **kwargs):
        pass

    @staticmethod
    def render(template_path, namespace):
        return json.dumps(namespace)


class GnocchiJinjaRenderer(templating.JinjaRenderer):
    def __init__(self, *args, **kwargs):
        super(GnocchiJinjaRenderer, self).__init__(*args, **kwargs)
        self.env.filters['tojson'] = json.dumps

    def render(self, template_path, namespace):
        namespace = dict(data=namespace)
        return super(GnocchiJinjaRenderer, self).render(
            template_path, namespace)


PECAN_CONFIG = {
    'app': {
        'root': 'gnocchi.rest.RootController',
        'modules': ['gnocchi.rest'],
    },
}


class NotImplementedMiddleware(object):
    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        try:
            return self.app(environ, start_response)
        except exceptions.NotImplementedError:
            raise webob.exc.HTTPNotImplemented(
                "Sorry, this Gnocchi server does "
                "not implement this feature 😞")


def setup_app(config=PECAN_CONFIG, cfg=None):
    if cfg is None:
        # NOTE(jd) That sucks but pecan forces us to use kwargs :(
        raise RuntimeError("Config is actually mandatory")
    s = config.get('storage')
    if not s:
        s = storage.get_driver(cfg)
    i = config.get('indexer')
    if not i:
        i = indexer.get_driver(cfg)
    i.connect()

    root_dir = os.path.dirname(os.path.abspath(__file__))

    # NOTE(sileht): pecan debug won't work in multi-process environment
    pecan_debug = cfg.api.pecan_debug
    if cfg.api.workers != 1 and pecan_debug:
        pecan_debug = False
        LOG.warning('pecan_debug cannot be enabled, if workers is > 1, '
                    'the value is overrided with False')

    app = pecan.make_app(
        config['app']['root'],
        debug=pecan_debug,
        hooks=(GnocchiHook(s, i, cfg),),
        guess_content_type_from_ext=False,
        custom_renderers={'json': OsloJSONRenderer,
                          'gnocchi_jinja': GnocchiJinjaRenderer},
        default_renderer='gnocchi_jinja',
        template_path=root_dir + "/templates",
    )

    app = wsgi.SharedDataMiddleware(
        app,
        {"/static": root_dir + "/static"},
        cache=not cfg.api.pecan_debug)

    if config.get('not_implemented_middleware', True):
        app = webob.exc.HTTPExceptionMiddleware(NotImplementedMiddleware(app))

    for middleware in reversed(cfg.api.middlewares):
        if not middleware:
            continue
        klass = importutils.import_class(middleware)
        # FIXME(jd) Find a way to remove that special handling…
        if klass == keystonemiddleware.auth_token.AuthProtocol:
            middleware_config = dict(cfg.keystone_authtoken)
        else:
            middleware_config = dict(cfg)
        app = klass(app, middleware_config)

    return app


def build_server():
    conf = service.prepare_service()
    serving.run_simple(conf.api.host, conf.api.port,
                       setup_app(cfg=conf),
                       processes=conf.api.workers)



#
# Copyright 2014 eNovance
#
# Authors: Mehdi Abaakouk <mehdi.abaakouk@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc

import six


@six.add_metaclass(abc.ABCMeta)
class ResourceBase(object):
    """Base class for resource."""

    @abc.abstractmethod
    def get_resource_extra_attributes(self, sample):
        """Extract the metadata from a ceilometer sample.

        :param sample: The ceilometer sample
        :returns: the resource attributes
        """

    @abc.abstractmethod
    def get_metrics_names(self):
        """Return the metric handled by this resource.

        :returns: list of metric names
        """



# -*- encoding: utf-8 -*-
#
# Copyright © 2015 eNovance
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import datetime
import uuid

import mock
from oslo_utils import timeutils

from gnocchi import statsd
from gnocchi import storage
from gnocchi.tests import base as tests_base


class TestStatsd(tests_base.TestCase):

    STATSD_USER_ID = uuid.uuid4()
    STATSD_PROJECT_ID = uuid.uuid4()
    STATSD_ARCHIVE_POLICY_NAME = "medium"

    def setUp(self):
        super(TestStatsd, self).setUp()

        self.conf.set_override("resource_id",
                               uuid.uuid4(), "statsd")
        self.conf.set_override("user_id",
                               self.STATSD_USER_ID, "statsd")
        self.conf.set_override("project_id",
                               self.STATSD_PROJECT_ID, "statsd")
        self.conf.set_override("archive_policy_name",
                               self.STATSD_ARCHIVE_POLICY_NAME, "statsd")

        # NOTE(jd) Always use self.server.storage and self.server.indexer to
        # pick at the right storage/indexer used by the statsd server, and not
        # new instances from the base test class.
        self.stats = statsd.Stats(self.conf)
        self.server = statsd.StatsdServer(self.stats)

    def test_flush_empty(self):
        self.server.stats.flush()

    @mock.patch.object(timeutils, 'utcnow')
    def _test_gauge_or_ms(self, metric_type, utcnow):
        metric_name = "test_gauge_or_ms"
        metric_key = metric_name + "|" + metric_type
        utcnow.return_value = datetime.datetime(2015, 1, 7, 13, 58, 36)
        self.server.datagram_received(
            ("%s:1|%s" % (metric_name, metric_type)).encode('ascii'),
            ("127.0.0.1", 12345))
        self.stats.flush()

        r = self.stats.indexer.get_resource('generic',
                                            self.conf.statsd.resource_id,
                                            with_metrics=True)

        measures = self.stats.storage.get_measures(storage.Metric(
            r.get_metric(metric_key), None))
        self.assertEqual([(datetime.datetime(2015, 1, 7), 86400.0, 1.0),
                          (datetime.datetime(2015, 1, 7, 13), 3600.0, 1.0),
                          (datetime.datetime(2015, 1, 7, 13, 58), 60.0, 1.0)],
                         measures)

        utcnow.return_value = datetime.datetime(2015, 1, 7, 13, 59, 37)
        # This one is going to be ignored
        self.server.datagram_received(
            ("%s:45|%s" % (metric_name, metric_type)).encode('ascii'),
            ("127.0.0.1", 12345))
        self.server.datagram_received(
            ("%s:2|%s" % (metric_name, metric_type)).encode('ascii'),
            ("127.0.0.1", 12345))
        self.stats.flush()

        measures = self.stats.storage.get_measures(storage.Metric(
            r.get_metric(metric_key), None))
        self.assertEqual([(datetime.datetime(2015, 1, 7), 86400.0, 1.5),
                          (datetime.datetime(2015, 1, 7, 13), 3600.0, 1.5),
                          (datetime.datetime(2015, 1, 7, 13, 58), 60.0, 1.0),
                          (datetime.datetime(2015, 1, 7, 13, 59), 60.0, 2.0)],
                         measures)

    def test_gauge(self):
        self._test_gauge_or_ms("g")

    def test_ms(self):
        self._test_gauge_or_ms("ms")

    @mock.patch.object(timeutils, 'utcnow')
    def test_counter(self, utcnow):
        metric_name = "test_counter"
        metric_key = metric_name + "|c"
        utcnow.return_value = datetime.datetime(2015, 1, 7, 13, 58, 36)
        self.server.datagram_received(
            ("%s:1|c" % metric_name).encode('ascii'),
            ("127.0.0.1", 12345))
        self.stats.flush()

        r = self.stats.indexer.get_resource('generic',
                                            self.conf.statsd.resource_id,
                                            with_metrics=True)

        measures = self.stats.storage.get_measures(storage.Metric(
            r.get_metric(metric_key), None))
        self.assertEqual([(datetime.datetime(2015, 1, 7), 86400.0, 1.0),
                          (datetime.datetime(2015, 1, 7, 13), 3600.0, 1.0),
                          (datetime.datetime(2015, 1, 7, 13, 58), 60.0, 1.0)],
                         measures)

        utcnow.return_value = datetime.datetime(2015, 1, 7, 13, 59, 37)
        self.server.datagram_received(
            ("%s:45|c" % metric_name).encode('ascii'),
            ("127.0.0.1", 12345))
        self.server.datagram_received(
            ("%s:2|c|@0.2" % metric_name).encode('ascii'),
            ("127.0.0.1", 12345))
        self.stats.flush()

        measures = self.stats.storage.get_measures(storage.Metric(
            r.get_metric(metric_key), None))
        self.assertEqual([(datetime.datetime(2015, 1, 7), 86400.0, 28),
                          (datetime.datetime(2015, 1, 7, 13), 3600.0, 28),
                          (datetime.datetime(2015, 1, 7, 13, 58), 60.0, 1.0),
                          (datetime.datetime(2015, 1, 7, 13, 59), 60.0, 55.0)],
                         measures)



#!/usr/bin/env python
#
# Copyright (c) 2014 eNovance
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#
# Tools to analyse the result of multiple call of duration_perf_test.py:
#
#   $ clients=10
#   $ parallel --progress -j $clients python duration_perf_test.py \
#       --result myresults/client{} ::: $(seq 0 $clients)
#   $ python duration_perf_analyse.py myresults
#    * get_measures:
#                  Time
#    count  1000.000000
#    mean      0.032090
#    std       0.028287
#    ...
#


import argparse
import os

import pandas


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('result',
                        help=('Path of the results of perf_tool.py.'),
                        default='result')

    data = {
        'get_measures': [],
        'write_measures': [],
        'write_metric': [],
    }
    args = parser.parse_args()
    for root, dirs, files in os.walk(args.result):
        for name in files:
            for method in data:
                if name.endswith('_%s.csv' % method):
                    datum = data[method]
                    filepath = os.path.join(root, name)
                    datum.append(pandas.read_csv(filepath))
                    cname = name.replace('_%s.csv' % method, '')
                    datum[-1].rename(columns={'Duration': cname}, inplace=True)

    for method in data:
        merged = pandas.datumFrame(columns=['Index', 'Duration'])
        append = pandas.datumFrame(columns=['Duration'])
        for datum in data[method]:
            datum.dropna(axis=1, inplace=True)
            datum.drop('Count', axis=1, inplace=True)
            merged = merged.merge(datum, on='Index')
            cname = datum.columns.values[1]
            datum.rename(columns={cname: 'Duration'}, inplace=True)
            append = append.append(datum.drop('Index', axis=1))
        merged.to_csv(os.path.join(args.result, '%s_merged.csv' % method),
                      index=False)
        print("* %s:" % method)
        print(append.describe())
        print("")

if __name__ == '__main__':
    main()
